--======================================================--
--                                                      --
--  NORTHEASTERN UNIVERSITY                             --
--  DEPARTMENT OF ELECTRICAL AND COMPUTER ENGINEERING   --
--  RAPID PROTOTYPING LABORATORY                        --
--                                                      --
--  FILE         | correction.vhd                       --
--  -------------+------------------------------------  --
--  DESCRIPTION  | Module to perform correction at      --
--               | the end of floating-point addition   --
--  -------------+------------------------------------  --
--  AUTHOR       | Pavle Belanovic                      --
--  -------------+------------------------------------  --
--  DATE         | 20 June 2002                         --
--                                                      --
--======================================================--

--******************************************************************************--
--                                                                              --
--	Copyright (C) 2000 Pavle Belanovic                                          --
--                                                                              --
--	This program is free software; you can redistribute it and/or               --
--	modify it under the terms of the GNU General Public License                 --
--	as published by the Free Software Foundation; either version 2              --
--	of the License, or (at your option) any later version.                      --
--                                                                              --
--	This program is distributed in the hope that it will be useful,             --
--	but WITHOUT ANY WARRANTY; without even the implied warranty of              --
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               --
--	GNU General Public License for more details.                                --
--                                                                              --
--	You should have received a copy of the GNU General Public License           --
--	along with this program; if not, write to the Free Software                 --
--	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. --
--                                                                              --
--******************************************************************************--

--======================================================--
--                      LIBRARIES                       --
--======================================================--

-- IEEE Libraries --
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

-- float
library work;
use work.float_pkg.all;

----------------------------------------------------------
--                  Correction Module                   --
----------------------------------------------------------
entity correction is
	generic
	(
		exp_bits			:	integer	:=	0;
		man_bits			:	integer	:=	0
	);
	port
	(
		--inputs
		F							:	in		std_logic_vector(man_bits downto 0);
		E							:	in		std_logic_vector(exp_bits-1 downto 0);
		S							:	in		std_logic;
		READY					:	in		std_logic;
		CLEAR					:	in		std_logic;
		ENABLE				:	in		std_logic;
		EXCEPTION_IN	:	in		std_logic;
		CLK						:	in		std_logic;
		--outputs
		RESULT				:	out		std_logic_vector(man_bits+exp_bits+1 downto 0)	:=	(others=>'0');
		EXCEPTION_OUT	:	out		std_logic																				:=	'0';
		DONE					:	out		std_logic																				:=	'0'
	);
end correction;

----------------------------------------------------------
--                  Correction Module                   --
----------------------------------------------------------
architecture correction_arch of correction is
	--SIGNALS
	signal	corrected_exponent	:	std_logic_vector(exp_bits-1 downto 0)	:=	(others=>'0');
	signal	corrected_mantissa	:	std_logic_vector(man_bits downto 0)		:=	(others=>'0');
	signal	one									:	std_logic_vector(exp_bits-1 downto 0)	:=	(others=>'0');
	signal	cout								:	std_logic															:=	'0';
	signal	clear_int						:	std_logic															:=	'0';
	signal	exc_int							:	std_logic															:=	'0';
	signal	zero								:	std_logic_vector(man_bits+exp_bits+1 downto 0)	:=	(others=>'0');
	signal	result_int					:	std_logic_vector(man_bits+exp_bits+1 downto 0)	:=	(others=>'0');
	signal	result_out					:	std_logic_vector(man_bits+exp_bits+1 downto 0)	:=	(others=>'0');
begin
	--ASYNCHRONOUS
	--component instantiation
	pa	:	parameterized_adder
		generic map
		(
			bits	=>	exp_bits
		)
		port map
		(
			A			=>	E,
			B			=>	one,
			CIN		=>	'0',
			S			=>	corrected_exponent,
			COUT	=>	cout
		);
	exp_mux	:	parameterized_mux
		generic map
		(
			bits	=>	exp_bits
		)
		port map
		(
			A			=>	corrected_exponent,
			B			=>	E,
			S			=>	ENABLE,
			O			=>	result_int(man_bits+exp_bits downto man_bits+1)
		);
	man_mux	:	parameterized_mux
		generic map
		(
			bits	=>	man_bits+1
		)
		port map
		(
			A			=>	corrected_mantissa,
			B			=>	F,
			S			=>	ENABLE,
			O			=>	result_int(man_bits downto 0)
		);
	out_mux	:	parameterized_mux
		generic map
		(
			bits	=>	exp_bits+man_bits+2
		)
		port map
		(
			A			=>	zero,
			B			=>	result_int,
			S			=>	clear_int,
			O			=>	result_out
		);
	--signal connections
	clear_int									<=	CLEAR OR exc_int;
	exc_int										<=	EXCEPTION_IN OR cout;
	zero											<=	(others=>'0');
	one(exp_bits-1 downto 1)	<=	(others=>'0');
	one(0)										<=	'1';
	corrected_mantissa(man_bits-1 downto 0)	<=	F(man_bits downto 1);
	corrected_mantissa(man_bits)						<=	'1';
	result_int(man_bits+exp_bits+1)					<=	S;
	--SYNCHRONOUS
	main	:	process(CLK)
	begin
		if(rising_edge(CLK)) then
			DONE					<=	READY;
			RESULT				<=	result_out;
			EXCEPTION_OUT	<=	exc_int;
		end if;--CLK
	end process;--main
end correction_arch; --end of architecture
