--======================================================--
--                                                      --
--  NORTHEASTERN UNIVERSITY                             --
--  DEPARTMENT OF ELECTRICAL AND COMPUTER ENGINEERING   --
--  RAPID PROTOTYPING LABORATORY                        --
--                                                      --
--  FILE         | float_pkg.vhd                        --
--  -------------+------------------------------------  --
--  DESCRIPTION  | Package for the parameterized        --
--               | floating-point library.              --
--  -------------+------------------------------------  --
--  AUTHOR       | Pavle Belanovic                      --
--  -------------+------------------------------------  --
--  DATE         | 20 June 2002                         --
--                                                      --
--======================================================--

--******************************************************************************--
--                                                                              --
--	Copyright (C) 2000 Pavle Belanovic                                          --
--                                                                              --
--	This program is free software; you can redistribute it and/or               --
--	modify it under the terms of the GNU General Public License                 --
--	as published by the Free Software Foundation; either version 2              --
--	of the License, or (at your option) any later version.                      --
--                                                                              --
--	This program is distributed in the hope that it will be useful,             --
--	but WITHOUT ANY WARRANTY; without even the implied warranty of              --
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               --
--	GNU General Public License for more details.                                --
--                                                                              --
--	You should have received a copy of the GNU General Public License           --
--	along with this program; if not, write to the Free Software                 --
--	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. --
--                                                                              --
--******************************************************************************--

--======================================================--
--                      LIBRARIES                       --
--======================================================--

-- IEEE Libraries --
library IEEE;
use IEEE.std_logic_1164.all;

-- LAD Mux Libraries --
--library LAD_Mux_Lib;
--use LAD_Mux_Lib.LAD_Mux_pkg.all;

-- Mem Mux Libraries --
--library Mem32_Mux_Lib, Mem64_Mux_Lib;
--use Mem32_Mux_Lib.Mem32_Mux_pkg.all;
--use Mem64_Mux_Lib.Mem64_Mux_pkg.all;

--======================================================--
--                       PACKAGE                        --
--======================================================--

package float_pkg is

----------------------------------------------------------
--                    Components                        --
----------------------------------------------------------
	--Round-normalize module
	component rnd_norm is
		generic
		(
			exp_bits			:	integer	:=	0;
			man_bits_in		:	integer	:=	0;
			man_bits_out	:	integer	:=	0
		);
		port
		(
			--inputs
			IN1						:	in		std_logic_vector((exp_bits+man_bits_in) downto 0);
			READY					:	in		std_logic;
			CLK						:	in		std_logic;
			ROUND					:	in		std_logic;
			EXCEPTION_IN	:	in		std_logic;
			--outputs
			OUT1					:	out		std_logic_vector((exp_bits+man_bits_out) downto 0)	:=	(others=>'0');
			DONE					:	out		std_logic																						:=	'0';
			EXCEPTION_OUT	:	out		std_logic																						:=	'0'
		);
	end component;

	--Floating point adder
	component fp_add is
		generic
		(
			exp_bits			:	integer	:=	0;
			man_bits			:	integer	:=	0
		);
		port
		(
			--inputs
			OP1						:	in		std_logic_vector(man_bits+exp_bits downto 0);
			OP2						:	in		std_logic_vector(man_bits+exp_bits downto 0);
			READY					:	in		std_logic;
			EXCEPTION_IN	:	in		std_logic;
			CLK						:	in		std_logic;
			--outputs
			RESULT				:	out		std_logic_vector(man_bits+exp_bits+1 downto 0)	:=	(others=>'0');
			EXCEPTION_OUT	:	out		std_logic																				:=	'0';
			DONE					:	out		std_logic																				:=	'0'
		);
	end component;

	--Rounding addition module
	component round_add is
		generic
		(
			man_bits_in		:	integer	:=	0;
			man_bits_out	:	integer	:=	0
		);
		port
		(
			--inputs
			MAN_IN				:	in		std_logic_vector(man_bits_in-1 downto 0);
			READY					:	in		std_logic;
			CLK						:	in		std_logic;
			ENABLE				:	in		std_logic;
			EXCEPTION_IN	:	in		std_logic;
			--outputs
			MAN_OUT				:	out		std_logic_vector(man_bits_out-1 downto 0)	:=	(others=>'0');
			EXCEPTION_OUT	:	out		std_logic																	:=	'0';
			DONE					:	out		std_logic																	:=	'0'
		);
	end component;
	
	--Normalizer
	component normalizer is
		generic
		(
			exp_bits			:	integer	:=	0;
			man_bits			:	integer	:=	0
		);
		port
		(
			--inputs
			MAN_IN				:	in		std_logic_vector(man_bits-1 downto 0);
			EXP_IN				:	in		std_logic_vector(exp_bits-1 downto 0);
			READY					:	in		std_logic;
			CLK						:	in		std_logic;
			EXCEPTION_IN	:	in		std_logic;
			--outputs
			MAN_OUT				:	out		std_logic_vector(man_bits-1 downto 0)	:=	(others=>'0');
			EXP_OUT				:	out		std_logic_vector(exp_bits-1 downto 0)	:=	(others=>'0');
			EXCEPTION_OUT	:	out		std_logic															:=	'0';
			DONE					:	out		std_logic															:=	'0'
		);
	end component;

	--Correction module
	component correction is
		generic
		(
			exp_bits			:	integer	:=	0;
			man_bits			:	integer	:=	0
		);
		port
		(
			--inputs
			F							:	in		std_logic_vector(man_bits downto 0);
			E							:	in		std_logic_vector(exp_bits-1 downto 0);
			S							:	in		std_logic;
			READY					:	in		std_logic;
			CLEAR					:	in		std_logic;
			ENABLE				:	in		std_logic;
			EXCEPTION_IN	:	in		std_logic;
			CLK						:	in		std_logic;
			--outputs
			RESULT				:	out		std_logic_vector(man_bits+exp_bits+1 downto 0)	:=	(others=>'0');
			EXCEPTION_OUT	:	out		std_logic																				:=	'0';
			DONE					:	out		std_logic																				:=	'0'
		);
	end component;
	
	--Add/subtract module
	component add_sub is
		generic
		(
			man_bits			:	integer	:=	0
		);
		port
		(
			--inputs
			F1						:	in		std_logic_vector(man_bits downto 0);
			F2						:	in		std_logic_vector(man_bits downto 0);
			READY					:	in		std_logic;
			OP						:	in		std_logic;
			CLK						:	in		std_logic;
			--outputs
			F_OUT					:	out		std_logic_vector(man_bits downto 0)	:=	(others=>'0');
			OVERFLOW			:	out		std_logic														:=	'0';
			DONE					:	out		std_logic														:=	'0'
		);
	end component;

	--Shift/adjust module
	component shift_adjust is
		generic
		(
			exp_bits			:	integer	:=	0;
			man_bits			:	integer	:=	0
		);
		port
		(
			--inputs
			E1						:	in		std_logic_vector(exp_bits-1 downto 0);
			E2						:	in		std_logic_vector(exp_bits-1 downto 0);
			F2						:	in		std_logic_vector(man_bits-1 downto 0);
			READY					:	in		std_logic;
			CLK						:	in		std_logic;
			--outputs
			E_OUT					:	out		std_logic_vector(exp_bits-1 downto 0)	:=	(others=>'0');
			F_OUT					:	out		std_logic_vector(man_bits downto 0)		:=	(others=>'0');
			DONE					:	out		std_logic 														:=	'0'
		);
	end component;

	--Swap module
	component swap is
		generic
		(
			exp_bits		:	integer	:=	0;
			man_bits		:	integer	:=	0
		);
		port
		(
			--inputs
			IN1						:	in		std_logic_vector((exp_bits+man_bits) downto 0);
			IN2						:	in		std_logic_vector((exp_bits+man_bits) downto 0);
			READY					:	in		std_logic;
			CLK						:	in		std_logic;
			--outputs
			OUT1					:	out		std_logic_vector((exp_bits+man_bits) downto 0)	:=	(others=>'0');
			OUT2					:	out		std_logic_vector((exp_bits+man_bits) downto 0)	:=	(others=>'0');
			DONE					:	out		std_logic																				:=	'0'
		);
	end component;

	--Parameterized priority encoder
	component parameterized_priority_encoder is
		generic
		(
			man_bits		:	integer	:=	0;
			shift_bits	:	integer	:=	0
		);
		port
		(
			--inputs
			MAN_IN				:	in		std_logic_vector(man_bits-1 downto 0);
			--outputs
			SHIFT					:	out		std_logic_vector(shift_bits-1 downto 0)	:=	(others=>'0');
			EXCEPTION_OUT	:	out	std_logic																	:=	'0'
		);
	end component;

	--Parameterized adder
	component parameterized_adder is
		generic
		(
			bits				:	integer						:=	0
		);
		port
		(
			--inputs
			A			:	in		std_logic_vector(bits-1 downto 0);
			B			:	in		std_logic_vector(bits-1 downto 0);
			CIN		:	in		std_logic;
			--outputs
			S			:	out		std_logic_vector(bits-1 downto 0)	:=	(others=>'0');
			COUT	:	out		std_logic													:=	'0'
		);
	end component;

	--Parameterized variable shifter	
	component parameterized_variable_shifter is
		generic
		(
			bits					:	integer		:=	0;
			select_bits		:	integer		:=	0;
			direction			:	std_logic	:=	'0' --0=right,1=left
		);
		port
		(
			--inputs
			I							:	in		std_logic_vector(bits-1 downto 0);
			S							:	in		std_logic_vector(select_bits-1 downto 0);
			CLEAR					:	in		std_logic;
			--outputs
			O							:	out		std_logic_vector(bits-1 downto 0)
		);
	end component;

	--Shifter control
	component shifter_control is
		generic
		(
			exp_bits			:	integer	:=	0;
			select_bits		:	integer	:=	0
		);
		port
		(
			--inputs
			DELTA					:	in		std_logic_vector(exp_bits-1 downto 0);
			--outputs
			CLEAR					:	out		std_logic;
			SHIFT					:	out		std_logic_vector(select_bits-1 downto 0)
		);
	end component;

	--Parameterized variable shifter row
	component parameterized_variable_shifter_row is
		generic
		(
			bits					:	integer		:=	0;
			row_number		:	integer		:=	0;
			direction			:	std_logic	:=	'0' --0=right,1=left
		);
		port
		(
			--inputs
			I							:	in		std_logic_vector(bits-1 downto 0);
			S							:	in		std_logic;
			--outputs
			O							:	out		std_logic_vector(bits-1 downto 0)
		);
	end component;

	--Parameterized comparator
	component parameterized_comparator is
		generic
		(
			bits				:	integer						:=	0
		);
		port
		(
			--inputs
			A						:	in		std_logic_vector(bits-1 downto 0);
			B						:	in		std_logic_vector(bits-1 downto 0);
			--outputs
			A_GT_B			:	out		std_logic		:=	'0';
			A_EQ_B			:	out		std_logic		:=	'0';
			A_LT_B			:	out		std_logic		:=	'0'
		);
	end component;

	--Parameterized multiplexer
	component parameterized_mux is
		generic
		(
			bits				:	integer						:=	0
		);
		port
		(
			--inputs
			A			:	in		std_logic_vector(bits-1 downto 0);
			B			:	in		std_logic_vector(bits-1 downto 0);
			S			:	in		std_logic;
			--outputs
			O			:	out		std_logic_vector(bits-1 downto 0)	:= (others=>'0')
		);
	end component;

	--Parameterized OR gate
	component parameterized_or_gate is
		generic
		(
			bits				:	integer						:=	0
		);
		port
		(
			--inputs
			A			:	in		std_logic_vector(bits-1 downto 0);
			--outputs
			O			:	out		std_logic	:=	'0'
		);
	end component;

	--Parameterized subtractor	
	component parameterized_subtractor is
		generic
		(
			bits				:	integer						:=	0
		);
		port
		(
			--inputs
			A			:	in		std_logic_vector(bits-1 downto 0);
			B			:	in		std_logic_vector(bits-1 downto 0);
			--outputs
			O			:	out		std_logic_vector(bits-1 downto 0)	:=	(others=>'0')
		);
	end component;

	--1-bit comparator
	component one_bit_comparator is
		port
		(
			--inputs
			A						:	in		std_logic;
			B						:	in		std_logic;
			GT					:	in		std_logic;
			EQ					:	in		std_logic;
			LT					:	in		std_logic;
			--outputs
			A_GT_B			:	out		std_logic		:=	'0';
			A_EQ_B			:	out		std_logic		:=	'0';
			A_LT_B			:	out		std_logic		:=	'0'
		);
	end component;

	--2 input multiplexer
	component mux2 is
		port
		(
			--inputs
			A			:	in		std_logic;
			B			:	in		std_logic;
			S			:	in		std_logic;
			--outputs
			O			:	out		std_logic		:=	'0'
		);
	end component;

	--Denormalize	
	component denorm is
		generic
		(
			exp_bits		:	integer	:=	0;
			man_bits		:	integer	:=	0
		);
		port
		(
			--inputs
			IN1						:	in		std_logic_vector(exp_bits+man_bits downto 0);
			READY					:	in		std_logic;
			EXCEPTION_IN	:	in		std_logic;
			--outputs
			OUT1					:	out		std_logic_vector(exp_bits+man_bits+1 downto 0)	:=	(others=>'0');
			DONE					:	out		std_logic																				:=	'0';
			EXCEPTION_OUT	:	out		std_logic																				:=	'0'
		);
	end component;

	--parameterized delay block
	component delay_block is
		generic
		(
			bits	:	integer	:=	0;
			delay	:	integer	:=	0
		);
		port
		(
			--inputs
			A					:	in	std_logic_vector(bits-1 downto 0);
			CLK				:	in	std_logic;
			--outputs
			A_DELAYED	:	out	std_logic_vector(bits-1 downto 0)	:=	(others=>'0')
		);
	end component;

	--parameterized absolute value
	component parameterized_absolute_value is
		generic
		(
			bits	:				integer															:=	0
		);
		port
		(
			--inputs
			IN1		:	in		std_logic_vector(bits-1 downto 0);
			--outputs
			EXC		:	out		std_logic														:=	'0';
			OUT1	:	out		std_logic_vector(bits-1 downto 0)		:=	(others=>'0')
		);
	end component;

	--fixed point to floating point conversion
	component fix2float is
		generic
		(
			fix_bits	:	integer	:=	0;
			exp_bits	:	integer	:=	0;
			man_bits	:	integer	:=	0
		);
		port
		(
			--inputs
			FIXED					:	in	std_logic_vector(fix_bits-1 downto 0);
			ROUND					:	in	std_logic;
			EXCEPTION_IN	:	in	std_logic;
			CLK						:	in	std_logic;
			READY					:	in	std_logic;
			--outputs
			FLOAT					:	out	std_logic_vector(exp_bits+man_bits downto 0)	:=	(others=>'0');
			EXCEPTION_OUT	:	out	std_logic																			:=	'0';
			DONE					:	out	std_logic																			:=	'0'
		);
	end component;

	--floating point to fixed point conversion
	component float2fix is
		generic
		(
			fix_bits	:	integer	:=	0;
			exp_bits	:	integer	:=	0;
			man_bits	:	integer	:=	0
		);
		port
		(
			--inputs
			FLOAT					:	in	std_logic_vector(exp_bits+man_bits downto 0);
			ROUND					:	in	std_logic;
			EXCEPTION_IN	:	in	std_logic;
			CLK						:	in	std_logic;
			READY					:	in	std_logic;
			--outputs
			FIXED					:	out	std_logic_vector(fix_bits-1 downto 0)					:=	(others=>'0');
			EXCEPTION_OUT	:	out	std_logic																			:=	'0';
			DONE					:	out	std_logic																			:=	'0'
		);
		end component;

	component fp_sub is
		generic
		(
			exp_bits			:	integer	:=	0;
			man_bits			:	integer	:=	0
		);
		port
		(
			--inputs
			OP1						:	in		std_logic_vector(man_bits+exp_bits downto 0);
			OP2						:	in		std_logic_vector(man_bits+exp_bits downto 0);
			READY					:	in		std_logic;
			EXCEPTION_IN	:	in		std_logic;
			CLK						:	in		std_logic;
			--outputs
			RESULT				:	out		std_logic_vector(man_bits+exp_bits+1 downto 0)	:=	(others=>'0');
			EXCEPTION_OUT	:	out		std_logic																				:=	'0';
			DONE					:	out		std_logic																				:=	'0'
		);
	end component;

	component parameterized_multiplier is
		generic
		(
			bits				:	integer						:=	0
		);
		port
		(
			--inputs
			A			:	in		std_logic_vector(bits-1 downto 0);
			B			:	in		std_logic_vector(bits-1 downto 0);
			--outputs
			S			:	out		std_logic_vector((2*bits)-1 downto 0)	:=	(others=>'0')
		);
	end component;

	component fp_mul is
		generic
		(
			exp_bits			:	integer	:=	0;
			man_bits			:	integer	:=	0
		);
		port
		(
			--inputs
			OP1						:	in		std_logic_vector(exp_bits+man_bits downto 0);
			OP2						:	in		std_logic_vector(exp_bits+man_bits downto 0);
			READY					:	in		std_logic;
			EXCEPTION_IN	:	in		std_logic;
			CLK						:	in		std_logic;
			--outputs
			RESULT				:	out		std_logic_vector(exp_bits+(2*man_bits) downto 0)	:=	(others=>'0');
			EXCEPTION_OUT	:	out		std_logic																					:=	'0';
			DONE					:	out		std_logic																					:=	'0'
		);
	end component;

end package;
