--======================================================--
--                                                      --
--  NORTHEASTERN UNIVERSITY                             --
--  DEPARTMENT OF ELECTRICAL AND COMPUTER ENGINEERING   --
--  RAPID PROTOTYPING LABORATORY                        --
--                                                      --
--  FILE         | shift_adjust.vhd                     --
--  -------------+------------------------------------  --
--  DESCRIPTION  | Module that lines up mantissas       --
--               | during addition.                     --
--  -------------+------------------------------------  --
--  AUTHOR       | Pavle Belanovic                      --
--  -------------+------------------------------------  --
--  DATE         | 20 June 2002                         --
--                                                      --
--======================================================--

--******************************************************************************--
--                                                                              --
--	Copyright (C) 2000 Pavle Belanovic                                          --
--                                                                              --
--	This program is free software; you can redistribute it and/or               --
--	modify it under the terms of the GNU General Public License                 --
--	as published by the Free Software Foundation; either version 2              --
--	of the License, or (at your option) any later version.                      --
--                                                                              --
--	This program is distributed in the hope that it will be useful,             --
--	but WITHOUT ANY WARRANTY; without even the implied warranty of              --
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               --
--	GNU General Public License for more details.                                --
--                                                                              --
--	You should have received a copy of the GNU General Public License           --
--	along with this program; if not, write to the Free Software                 --
--	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. --
--                                                                              --
--******************************************************************************--

--======================================================--
--                      LIBRARIES                       --
--======================================================--

-- IEEE Libraries --
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

-- float
library work;
use work.float_pkg.all;

----------------------------------------------------------
--                Shift/Adjust Module                   --
----------------------------------------------------------
entity shift_adjust is
	generic
	(
		exp_bits			:	integer	:=	8;
		man_bits			:	integer	:=	23
	);
	port
	(
		--inputs
		E1						:	in		std_logic_vector(exp_bits-1 downto 0);
		E2						:	in		std_logic_vector(exp_bits-1 downto 0);
		F2						:	in		std_logic_vector(man_bits-1 downto 0);
		READY					:	in		std_logic;
		CLK						:	in		std_logic;
		--outputs
		E_OUT					:	out		std_logic_vector(exp_bits-1 downto 0)	:=	(others=>'0');
		F_OUT					:	out		std_logic_vector(man_bits downto 0)		:=	(others=>'0');
		DONE					:	out		std_logic 														:=	'0'
	);
end shift_adjust;

----------------------------------------------------------
--                Shift/Adjust Module                   --
----------------------------------------------------------
architecture shift_adjust_arch of shift_adjust is
	--FUNCTIONS
	function ceil_log2 (input: integer) return integer is
		variable max		:	integer :=1; -- exp=0 => max=2^0=1
		variable exp		:	integer :=0;
	begin
		log_loop: while max < input loop
			exp	:=	exp+1;	-- increment exponent
			max	:=	max*2;	-- max = 2^exp
		end loop log_loop;
		return exp; -- exponent that gives (2^exp) >= input
	end ceil_log2;
	--CONSTANTS
	constant	shift_width	:	integer	:=	ceil_log2(man_bits+1);
	--SIGNALS
	signal	delta		:	std_logic_vector(exp_bits-1 downto 0)			:=	(others=>'0');
	signal	clear		:	std_logic																	:=	'0';
	signal	shift		:	std_logic_vector(shift_width-1 downto 0)	:=	(others=>'0');
	signal	man_in	:	std_logic_vector(man_bits downto 0)				:=	(others=>'0');
	signal	man_out	:	std_logic_vector(man_bits downto 0)				:=	(others=>'0');
begin
	--ASYNCHRONOUS
	--component instantiation
	ps	:	parameterized_subtractor
		generic map
		(
			bits	=>	exp_bits
		)
		port map
		(
			A			=>	E1,
			B			=>	E2,
			O			=>	delta
		);
	sc	:	shifter_control
		generic map
		(
			exp_bits			=>	exp_bits,
			select_bits		=>	shift_width
		)
		port map
		(
			DELTA					=>	delta,
			CLEAR					=>	clear,
			SHIFT					=>	shift
		);
	pvs	:	parameterized_variable_shifter
		generic map
		(
			bits					=>	man_bits+1,
			select_bits		=>	shift_width,
			direction			=>	'0'
		)
		port map
		(
			I							=>	man_in,
			S							=>	shift,
			CLEAR					=>	clear,
			O							=>	man_out
		);
	--signal renaming
	man_in(man_bits downto 1)	<=	F2;
	man_in(0)									<=	'0';
	--SYNCHRONOUS
	main: process (CLK)
	begin
		if(rising_edge(CLK)) then
			DONE	<=	READY;
			E_OUT	<=	E1;
			F_OUT	<=	man_out;
		end if;--CLK
	end process;--main
end shift_adjust_arch; --end of architecture

